goog.addDependency('/input/base.js', ['goog'], []);
goog.addDependency('/input/simple.soy', ['testapp.simple'], ['soy', 'soy.StringBuilder']);
goog.addDependency('/input/soyutils_usegoog.js', ['soy', 'soy.StringBuilder'], ['goog.dom', 'goog.dom.NodeType', 'goog.format', 'goog.i18n.BidiFormatter', 'goog.i18n.bidi', 'goog.string', 'goog.string.StringBuffer']);
goog.addDependency('/input/test1.js', ['testApplication'], ['testapp']);
goog.addDependency('/input/test1.soy', ['testapp'], ['soy', 'soy.StringBuilder']);
goog.addDependency('/input/test2.js', ['testApplication2'], ['testapp']);
